@@ -7,12 +7,13 @@ Reads tool, resource, and prompt permission files and provides a singleton inter
 
 import json
 from dataclasses import dataclass
+from functools import cache
 from pathlib import Path
 from typing import Any
 
-from loguru import logger as log
+from loguru import logger as log  # pyright: ignore[reportUnknownVariableType]
 
-from src.config import Config, get_config_dir
+from src.config import Config, ensure_permissions_file, get_config_dir
 
 
 def _default_permissions_dir() -> Path:
@@ -52,6 +53,9 @@ class ResourcePermission:
     write_operation: bool = False
     read_private_data: bool = False
     read_untrusted_public_data: bool = False
+    acl: str = "PUBLIC"
+    # Optional metadata fields (ignored by enforcement but accepted from JSON)
+    description: str | None = None
 
 
 @dataclass
@@ -158,6 +162,12 @@ class Permissions:
             )
 
     @classmethod
+    def clear_permissions_file_cache(cls) -> None:
+        """Clear the cache for the JSON permissions files"""
+        cls._load_permission_file.cache_clear()
+
+    @classmethod
+    @cache
     def _load_permission_file(
         cls,
         file_path: Path,
@@ -170,8 +180,8 @@ class Permissions:
         permissions: dict[str, Any] = {}
         metadata: PermissionsMetadata | None = None
 
-        if not file_path.exists():
-            raise PermissionsError(f"Permissions file not found at {file_path}")
+        # Centralized bootstrap: ensure file exists or create from repo defaults
+        file_path = ensure_permissions_file(file_path)
 
         with open(file_path) as f:
             data: dict[str, Any] = json.load(f)
@@ -203,43 +213,117 @@ class Permissions:
     def get_tool_permission(self, tool_name: str) -> ToolPermission:
         """Get permission for a specific tool"""
         if tool_name not in self.tool_permissions:
-            raise PermissionsError(f"Tool '{tool_name}' not found in permissions")
+            if tool_name.startswith("builtin_"):
+                log.info(
+                    f"Tool '{tool_name}' not found; returning builtin safe default (enabled, 0 risk)"
+                )
+                return ToolPermission(
+                    enabled=True,
+                    write_operation=False,
+                    read_private_data=False,
+                    read_untrusted_public_data=False,
+                    acl="PUBLIC",
+                )
+            log.warning(
+                f"Tool '{tool_name}' not found in permissions; returning enabled full-trifecta default"
+            )
+            return ToolPermission(
+                enabled=True,
+                write_operation=True,
+                read_private_data=True,
+                read_untrusted_public_data=True,
+                acl="SECRET",
+            )
         return self.tool_permissions[tool_name]
 
     def get_resource_permission(self, resource_name: str) -> ResourcePermission:
         """Get permission for a specific resource"""
         if resource_name not in self.resource_permissions:
-            raise PermissionsError(f"Resource '{resource_name}' not found in permissions")
+            if resource_name.startswith("builtin_"):
+                log.info(
+                    f"Resource '{resource_name}' not found; returning builtin safe default (enabled, 0 risk)"
+                )
+                return ResourcePermission(
+                    enabled=True,
+                    write_operation=False,
+                    read_private_data=False,
+                    read_untrusted_public_data=False,
+                )
+            log.warning(
+                f"Resource '{resource_name}' not found in permissions; returning enabled full-trifecta default"
+            )
+            return ResourcePermission(
+                enabled=True,
+                write_operation=True,
+                read_private_data=True,
+                read_untrusted_public_data=True,
+            )
         return self.resource_permissions[resource_name]
 
     def get_prompt_permission(self, prompt_name: str) -> PromptPermission:
         """Get permission for a specific prompt"""
         if prompt_name not in self.prompt_permissions:
-            raise PermissionsError(f"Prompt '{prompt_name}' not found in permissions")
+            if prompt_name.startswith("builtin_"):
+                log.info(
+                    f"Prompt '{prompt_name}' not found; returning builtin safe default (enabled, 0 risk)"
+                )
+                return PromptPermission(
+                    enabled=True,
+                    write_operation=False,
+                    read_private_data=False,
+                    read_untrusted_public_data=False,
+                    acl="PUBLIC",
+                )
+            log.warning(
+                f"Prompt '{prompt_name}' not found in permissions; returning enabled full-trifecta default"
+            )
+            return PromptPermission(
+                enabled=True,
+                write_operation=True,
+                read_private_data=True,
+                read_untrusted_public_data=True,
+                acl="SECRET",
+            )
         return self.prompt_permissions[prompt_name]
 
     def is_tool_enabled(self, tool_name: str) -> bool:
         """Check if a tool is enabled
         Also checks if the server is enabled"""
         permission = self.get_tool_permission(tool_name)
-        server_name = self.server_name_from_tool_name(tool_name)
-        server_enabled = self.is_server_enabled(server_name)
+        try:
+            server_name = self.server_name_from_tool_name(tool_name)
+            server_enabled = self.is_server_enabled(server_name)
+        except PermissionsError:
+            log.warning(f"Server resolution failed for tool '{tool_name}'; treating as disabled")
+            server_enabled = False
         return permission.enabled and server_enabled
 
     def is_resource_enabled(self, resource_name: str) -> bool:
         """Check if a resource is enabled
         Also checks if the server is enabled"""
         permission = self.get_resource_permission(resource_name)
-        server_name = self.server_name_from_tool_name(resource_name)
-        server_enabled = self.is_server_enabled(server_name)
+        try:
+            server_name = self.server_name_from_tool_name(resource_name)
+            server_enabled = self.is_server_enabled(server_name)
+        except PermissionsError:
+            log.warning(
+                f"Server resolution failed for resource '{resource_name}'; treating as disabled"
+            )
+            server_enabled = False
         return permission.enabled and server_enabled
 
     def is_prompt_enabled(self, prompt_name: str) -> bool:
         """Check if a prompt is enabled
         Also checks if the server is enabled"""
         permission = self.get_prompt_permission(prompt_name)
-        server_name = self.server_name_from_tool_name(prompt_name)
-        server_enabled = self.is_server_enabled(server_name)
+        try:
+            server_name = self.server_name_from_tool_name(prompt_name)
+            server_enabled = self.is_server_enabled(server_name)
+        except PermissionsError:
+            log.warning(
+                f"Server resolution failed for prompt '{prompt_name}'; treating as disabled"
+            )
+            server_enabled = False
         return permission.enabled and server_enabled
 
     @staticmethod
