@@ -5,11 +5,21 @@ FastMCP instance for the single-user Open Edison setup.
 Handles MCP protocol communication with running servers using a unified composite proxy.
 """
 
+import asyncio
+import time
 from typing import Any, TypedDict
 
 from fastmcp import Client as FastMCPClient
 from fastmcp import Context, FastMCP
+from fastmcp.server.server import add_resource_prefix, has_resource_prefix
+
+# Low level FastMCP imports
+from fastmcp.tools.tool import Tool
+from fastmcp.tools.tool_transform import (
+    apply_transformations_to_tools,
+)
 from loguru import logger as log
+from mcp.server.lowlevel.server import LifespanResultT
 
 from src.config import Config, MCPServerConfig
 from src.middleware.session_tracking import (
@@ -60,6 +70,118 @@ class SingleUserMCP(FastMCP[Any]):
         self._setup_demo_resources()
         self._setup_demo_prompts()
 
+    async def import_server(
+        self,
+        server: FastMCP[LifespanResultT],
+        prefix: str | None = None,
+        tool_separator: str | None = None,
+        resource_separator: str | None = None,
+        prompt_separator: str | None = None,
+    ) -> None:
+        """
+        Import the MCP objects from another FastMCP server into this one with a given prefix.
+        Overloads FastMCP's import_server method to improve performance.
+
+        Args:
+            server: The FastMCP server to import
+            prefix: prefix to use for the imported server's objects. If None,
+                objects are imported with their original names.
+            tool_separator: Deprecated. Required to be None.
+            resource_separator: Deprecated. Required to be None.
+            prompt_separator: Deprecated. Required to be None.
+        """
+
+        if prefix is None:
+            raise ValueError("Prefix is required")
+
+        if tool_separator is not None:
+            raise ValueError("Tool separator is deprecated and not supported")
+
+        if resource_separator is not None:
+            raise ValueError("Resource separator is deprecated and not supported")
+
+        if prompt_separator is not None:
+            raise ValueError("Prompt separator is deprecated and not supported")
+
+        log.debug(f"🔧 Importing server {prefix} ({server.name}) into single user MCP'")
+
+        # Fetch all server objects in parallel
+        tools, resources, templates, prompts = await asyncio.gather(
+            server.get_tools(),
+            server.get_resources(),
+            server.get_resource_templates(),
+            server.get_prompts(),
+            return_exceptions=True,
+        )
+
+        # Validate and normalize all results
+        tools = self._validate_server_result(tools, "tools", prefix)
+        resources = self._validate_server_result(resources, "resources", prefix)
+        templates = self._validate_server_result(templates, "templates", prefix)
+        prompts = self._validate_server_result(prompts, "prompts", prefix)
+
+        # Import all components
+        self._import_tools(tools, prefix)
+        self._import_resources(resources, prefix)
+        self._import_templates(templates, prefix)
+        self._import_prompts(prompts, prefix)
+
+        log.debug(f"Imported server {server.name} with prefix '{prefix}'")
+
+    def _validate_server_result(
+        self, result: Any, result_type: str, server_name: str
+    ) -> dict[str, Any]:
+        """Validate and normalize server result from asyncio.gather with return_exceptions=True."""
+        if isinstance(result, RuntimeError):
+            log.debug(f'Server {server_name} does not appear to contain "{result_type}"')
+            return {}
+        if isinstance(result, Exception):
+            log.error(
+                f'Server {server_name} received and unexpected exception when feetching "{result_type}". result: {result}'
+            )
+            return {}
+        if not isinstance(result, dict):
+            log.warning(f"Server {server_name} returned an unexpected response")
+            log.debug(
+                f"Server {server_name} _validate_server_result unexpected type {type(result)} with value: {result}"
+            )
+            return {}
+        return result  # type: ignore[return-value]
+
+    def _import_tools(self, tools: dict[str, Any], prefix: str) -> None:
+        """Import tools from server"""
+        for key, tool in tools.items():
+            if prefix:
+                tool = tool.model_copy(key=f"{prefix}_{key}")
+            self._tool_manager.add_tool(tool)
+
+    def _import_resources(self, resources: dict[str, Any], prefix: str) -> None:
+        """Import resources from server"""
+        for key, resource in resources.items():
+            if prefix:
+                resource_key = add_resource_prefix(key, prefix, self.resource_prefix_format)
+                resource = resource.model_copy(
+                    update={"name": f"{prefix}_{resource.name}"}, key=resource_key
+                )
+            self._resource_manager.add_resource(resource)
+
+    def _import_templates(self, templates: dict[str, Any], prefix: str) -> None:
+        """Import templates from server"""
+        for key, template in templates.items():
+            if prefix:
+                template_key = add_resource_prefix(key, prefix, self.resource_prefix_format)
+                template = template.model_copy(
+                    update={"name": f"{prefix}_{template.name}"}, key=template_key
+                )
+            self._resource_manager.add_template(template)
+
+    def _import_prompts(self, prompts: dict[str, Any], prefix: str) -> None:
+        """Import prompts from server"""
+        for key, prompt in prompts.items():
+            if prefix:
+                prompt = prompt.model_copy(key=f"{prefix}_{key}")
+            self._prompt_manager.add_prompt(prompt)
+
     def _convert_to_fastmcp_config(self, enabled_servers: list[MCPServerConfig]) -> dict[str, Any]:
         """
         Convert Open Edison config format to FastMCP MCPConfig format.
@@ -87,46 +209,6 @@ class SingleUserMCP(FastMCP[Any]):
 
         return {"mcpServers": mcp_servers}
 
-    async def create_composite_proxy(self, enabled_servers: list[MCPServerConfig]) -> bool:
-        """
-        Create a unified composite proxy for all enabled MCP servers.
-
-        This replaces individual server mounting with a single FastMCP composite proxy
-        that handles all configured servers with automatic namespacing.
-
-        Args:
-            enabled_servers: List of enabled MCP server configurations
-
-        Returns:
-            True if composite proxy was created successfully, False otherwise
-        """
-        if not enabled_servers:
-            log.info("No real servers to mount in composite proxy")
-            return True
-
-        oauth_manager = get_oauth_manager()
-
-        for server_config in enabled_servers:
-            server_name = server_config.name
-
-            # Skip if this server would produce an empty config (e.g., misconfigured)
-            fastmcp_config = self._convert_to_fastmcp_config([server_config])
-            if not fastmcp_config.get("mcpServers"):
-                log.warning(f"Skipping server '{server_name}' due to empty MCP config")
-                continue
-
-            try:
-                await self._mount_single_server(server_config, fastmcp_config, oauth_manager)
-            except Exception as e:
-                log.error(f"❌ Failed to mount server {server_name}: {e}")
-                # Continue with other servers even if one fails
-                continue
-
-        log.info(
-            f"✅ Created composite proxy with {len(enabled_servers)} servers ({mounted_servers.keys()})"
-        )
-        return True
-
     async def _mount_single_server(
         self,
         server_config: MCPServerConfig,
@@ -140,6 +222,7 @@ class SingleUserMCP(FastMCP[Any]):
         remote_url = server_config.get_remote_url()
         oauth_info = await oauth_manager.check_oauth_requirement(server_name, remote_url)
 
+        client_timeout = 10
         # Create proxy based on server type to avoid union type issues
         if server_config.is_remote_server():
             # Handle remote servers (with or without OAuth)
@@ -156,18 +239,22 @@ class SingleUserMCP(FastMCP[Any]):
                     server_config.oauth_client_name,
                 )
                 if oauth_auth:
-                    client = FastMCPClient(remote_url, auth=oauth_auth)
+                    client = FastMCPClient(
+                        remote_url,
+                        auth=oauth_auth,
+                        timeout=client_timeout,
+                    )
                     log.info(
                         f"🔐 Created remote client with OAuth authentication for {server_name}"
                     )
                 else:
-                    client = FastMCPClient(remote_url)
+                    client = FastMCPClient(remote_url, timeout=client_timeout)
                     log.warning(
                         f"⚠️ OAuth auth creation failed, using unauthenticated client for {server_name}"
                     )
             else:
                 # Remote server without OAuth or needs auth
-                client = FastMCPClient(remote_url)
+                client = FastMCPClient(remote_url, timeout=client_timeout)
                 log.info(f"🌐 Created remote client for {server_name}")
 
             # Log OAuth status warnings
@@ -184,10 +271,12 @@ class SingleUserMCP(FastMCP[Any]):
 
         else:
             # Local server - create proxy directly from config (avoids union type issue)
-            log.info(f"🔧 Creating local process proxy for {server_name}")
+            log.debug(f"🔧 Creating local process proxy for {server_name}")
             proxy = FastMCP.as_proxy(fastmcp_config)
 
-        super().mount(proxy, prefix=server_name)
+        log.debug(f"🔧 Importing server {server_name} into single user MCP")
+        await self.import_server(proxy, prefix=server_name)
+        # await super().import_server(proxy, prefix=server_name)
         mounted_servers[server_name] = MountedServerInfo(config=server_config, proxy=proxy)
 
         server_type = "remote" if server_config.is_remote_server() else "local"
@@ -230,10 +319,7 @@ class SingleUserMCP(FastMCP[Any]):
         try:
             oauth_manager = get_oauth_manager()
             await self._mount_single_server(server_config, fastmcp_config, oauth_manager)
-            # Warm lists after mount
-            _ = await self._tool_manager.list_tools()
-            _ = await self._resource_manager.list_resources()
-            _ = await self._prompt_manager.list_prompts()
+
             return True
         except Exception as e:  # noqa: BLE001
             log.error(f"❌ Failed to mount server {server_name}: {e}")
@@ -242,7 +328,6 @@ class SingleUserMCP(FastMCP[Any]):
     async def unmount(self, server_name: str) -> bool:
         """
         Unmount a previously mounted server by name.
-
         Returns True if it was unmounted, False if it wasn't mounted.
         """
         info = mounted_servers.pop(server_name, None)
@@ -250,85 +335,145 @@ class SingleUserMCP(FastMCP[Any]):
             log.info(f"ℹ️  Server {server_name} was not mounted")
             return False
 
-        proxy = info.get("proxy")
-
-        # Manually remove from FastMCP managers' mounted lists
+        # Remove the server from mounted_servers lists in all managers
         for manager_name in ("_tool_manager", "_resource_manager", "_prompt_manager"):
             manager = getattr(self, manager_name, None)
+            if manager is None:
+                continue
             mounted_list = getattr(manager, "_mounted_servers", None)
             if mounted_list is None:
                 continue
 
-            # Prefer removing by both prefix and object identity; fallback to prefix-only
-            new_list = [
-                m
-                for m in mounted_list
-                if not (m.prefix == server_name and (proxy is None or m.server is proxy))
-            ]
-            if len(new_list) == len(mounted_list):
-                new_list = [m for m in mounted_list if m.prefix != server_name]
+            # Remove servers with matching prefix
+            mounted_list[:] = [m for m in mounted_list if m.prefix != server_name]
 
-            mounted_list[:] = new_list
+        # Remove tools with matching prefix (server name)
+        self._tool_manager._tools = {  # type: ignore
+            key: value
+            for key, value in self._tool_manager._tools.items()  # type: ignore
+            if not key.startswith(f"{server_name}_")
+        }
 
-        # Invalidate and warm lists to ensure reload
-        _ = await self._tool_manager.list_tools()
-        _ = await self._resource_manager.list_resources()
-        _ = await self._prompt_manager.list_prompts()
+        # Remove transformations with matching prefix (server name)
+        self._tool_manager.transformations = {  # type: ignore
+            key: value
+            for key, value in self._tool_manager.transformations.items()  # type: ignore
+            if not key.startswith(f"{server_name}_")
+        }
+
+        # Remove resources with matching prefix (server name)
+        self._resource_manager._resources = {  # type: ignore
+            key: value
+            for key, value in self._resource_manager._resources.items()  # type: ignore
+            if not has_resource_prefix(key, server_name, self.resource_prefix_format)  # type: ignore
+        }
+
+        # Remove templates with matching prefix (server name)
+        self._resource_manager._templates = {  # type: ignore
+            key: value
+            for key, value in self._resource_manager._templates.items()  # type: ignore
+            if not has_resource_prefix(key, server_name, self.resource_prefix_format)  # type: ignore
+        }
+
+        # Remove prompts with matching prefix (server name)
+        self._prompt_manager._prompts = {  # type: ignore
+            key: value
+            for key, value in self._prompt_manager._prompts.items()  # type: ignore
+            if not key.startswith(f"{server_name}_")
+        }
 
         log.info(f"🧹 Unmounted server {server_name} and cleared references")
         return True
 
-    async def _send_list_changed_notifications(self) -> None:
-        """Send notifications to clients about changed component lists."""
-        try:
-            # Import here to avoid circular imports
-            from fastmcp.server.dependencies import get_context
+    async def list_all_servers_tools_parallel(self) -> list[Tool]:
+        """Reload all servers' tools in parallel.
+        Reimplements FastMCP's ToolManager._list_tools method with parallel execution.
+        """
 
-            try:
-                context = get_context()
-                # Queue notifications for all component types since we don't know
-                # what types of components the unmounted server provided
-                context._queue_tool_list_changed()  # type: ignore
-                context._queue_resource_list_changed()  # type: ignore
-                context._queue_prompt_list_changed()  # type: ignore
-                log.debug("Queued component list change notifications")
-            except RuntimeError:
-                # No active context - notifications will be sent when context becomes available
-                log.debug("No active context for notifications")
+        # Execute all server reloads in parallel
+        list_tasks = [
+            server.server._list_tools()
+            for server in self._tool_manager._mounted_servers  # type: ignore
+        ]
 
-        except Exception as e:
-            log.warning(f"Error sending unmount notifications: {e}")
+        log.debug(f"Starting reload for {len(list_tasks)} servers' tools in parallel")
+        start_time = time.perf_counter()
+        all_tools: dict[str, Tool] = {}
+        if list_tasks:
+            # Use return_exceptions=True to prevent one failing server from breaking everything
+            tools_lists = await asyncio.gather(*list_tasks, return_exceptions=True)
+            for server, tools_result in zip(
+                self._tool_manager._mounted_servers,  # type: ignore
+                tools_lists,
+                strict=False,
+            ):
+                if isinstance(tools_result, Exception):
+                    log.warning(f"Failed to get tools from server {server.prefix}: {tools_result}")
+                    continue
+
+                tools_list = tools_result
+                if not tools_list or not isinstance(tools_list, list):
+                    continue
+
+                tools_dict = {t.key: t for t in tools_list}  # type: ignore
+                if server.prefix:
+                    for tool in tools_dict.values():
+                        prefixed_tool = tool.model_copy(  # type: ignore
+                            key=f"{server.prefix}_{tool.key}"  # type: ignore
+                        )
+                        all_tools[prefixed_tool.key] = prefixed_tool  # type: ignore
+                else:
+                    all_tools.update(tools_dict)  # type: ignore
+            log.debug(
+                f"Saved {len(all_tools)} tools from {len([r for r in tools_lists if not isinstance(r, Exception)])} servers"
+            )
+        else:
+            all_tools = {}
+
+        # Add local tools
+        all_tools.update(self._tool_manager._tools)  # type: ignore
+
+        transformed_tools = apply_transformations_to_tools(
+            tools=all_tools,
+            transformations=self._tool_manager.transformations,
+        )
+
+        final_tools_list = list(transformed_tools.values())
+
+        end_time = time.perf_counter()
+        log.debug(f"Time taken to reload all servers' tools: {end_time - start_time:.1f} seconds")
+        return final_tools_list
 
     async def initialize(self) -> None:
         """Initialize the FastMCP server using unified composite proxy approach."""
         log.info("Initializing Single User MCP server with composite proxy")
         log.debug(f"Available MCP servers in config: {[s.name for s in Config().mcp_servers]}")
-
+        start_time = time.perf_counter()
         # Get all enabled servers
         enabled_servers = [s for s in Config().mcp_servers if s.enabled]
         log.info(
             f"Found {len(enabled_servers)} enabled servers: {[s.name for s in enabled_servers]}"
         )
 
-        # Unmount all servers
-        for server_name in list(mounted_servers.keys()):
+        # Figure out which servers are to be unmounted
+        enabled_server_names = {s.name for s in enabled_servers}
+        servers_to_unmount = [s for s in mounted_servers if s not in enabled_server_names]
+
+        # Figure out which servers are to be mounted
+        servers_to_mount = [s.name for s in enabled_servers if s.name not in mounted_servers]
+
+        # Unmount those servers (quick)
+        for server_name in servers_to_unmount:
             await self.unmount(server_name)
 
-        # Create composite proxy for all real servers
-        success = await self.create_composite_proxy(enabled_servers)
-        if not success:
-            log.error("Failed to create composite proxy")
-            return
+        # Mount those servers (async gathered bc import does network roundtrip)
+        mount_tasks = [self.mount_server(server_name) for server_name in servers_to_mount]
+        await asyncio.gather(*mount_tasks)
 
         log.info("✅ Single User MCP server initialized with composite proxy")
-
-        # Invalidate and warm lists to ensure reload
-        _ = await self._tool_manager.list_tools()
-        _ = await self._resource_manager.list_resources()
-        _ = await self._prompt_manager.list_prompts()
-
-        # Send notifications to clients about changed component lists
-        await self._send_list_changed_notifications()
+        log.debug(
+            f"Time taken to initialize Single User MCP server: {time.perf_counter() - start_time:.1f} seconds"
+        )
 
     def _calculate_risk_level(self, trifecta: dict[str, bool]) -> str:
         """
@@ -418,7 +563,7 @@ class SingleUserMCP(FastMCP[Any]):
             """
             Get a list of all available tools. Use this tool to get an updated list of available tools.
             """
-            tool_list = await self._tool_manager.list_tools()
+            tool_list = await self.list_all_servers_tools_parallel()
             available_tools: list[str] = []
             log.trace(f"Raw tool list: {tool_list}")
             perms = Permissions()
@@ -454,7 +599,7 @@ class SingleUserMCP(FastMCP[Any]):
     def _setup_demo_resources(self) -> None:
         """Set up built-in demo resources for testing."""
 
-        @self.resource("config://app")  # noqa
+        @self.resource("info://builtin/app")  # noqa
         def builtin_get_app_config() -> dict[str, Any]:
             """Get application configuration."""
             return {
@@ -463,7 +608,7 @@ class SingleUserMCP(FastMCP[Any]):
                 "total_mounted": len(mounted_servers),
             }
 
-        log.info("✅ Added built-in demo resources: config://app")
+        log.info("✅ Added built-in demo resources: info://builtin/app")
 
     def _setup_demo_prompts(self) -> None:
         """Set up built-in demo prompts for testing."""
